/* 
 * Network accounting
 * capture.c - capture raw packets - pcap version
 * (C) 1996 Ulrich Callmeier
 * rewritten experimental pcap interface to netacct
 * to work with linux, bsd ... 
 * geroy
 * added missing ntohs() to srcport, dstport
 * fix to listen to more than 1 interface .. some ideas taken
 * from snort/ntop source
 */

#include "netacct.h"

#ifdef HAVE_LIBPCAP
#include <pcap.h>
#endif

#ifdef _Solaris_
/*#include <net/if.h>*/
#include <netinet/if_ether.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/ip_var.h>
#include <netinet/tcpip.h>
#else
#ifndef _OpenBSD_
#include <net/ethernet.h>
#endif
#endif
#if defined(_FreeBSD_) || defined(_OpenBSD_) 
#include <machine/param.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#endif

#ifdef _FreeBSD_
#include <netinet/ip_compat.h>
#endif

#ifdef _OpenBSD_
#include <netinet/if_ether.h>
#endif

#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


void handle_ip(unsigned char buf[], char *devname, char *user);
int *taskids[MAX_INTERFACES];
char perrbuff[PCAP_ERRBUF_SIZE];
#define PCAP_SNAPLEN 128
#define PCAP_TMOUT 1000

void handle_frame (unsigned char[], int);

/* prepace interfaces */
void init_capture()
{
  struct promisc_device *p;
  
  p = cfg -> promisc;
  
  while(p!=NULL)
    {
      pds[interface_number] = pcap_open_live(p -> name, PCAP_SNAPLEN, cfg->sniff, PCAP_TMOUT, perrbuff);

      //syslog(LOG_DEBUG, "pcap_open_live -> %s", p -> name);
      
      if(pds[interface_number] == NULL)
	{
	  syslog(LOG_ERR, "can't pcap_open_live: %s\n", perrbuff);
	  daemon_stop(0);
	}
      
      //syslog(LOG_DEBUG, "init_capture(), interface_number = %i, device = %s", interface_number, p->name);
      interface_number++;
      p = p -> next; /* next interface to set in promisc mode */
    }
  
  DEBUG(DBG_MISC, sprintf(dbg, "%s set to promiscous mode\n", p -> name));
}

/* close all promisc interfaces */
void exit_capture(void)
{
	int i = 0;
	while(pds[i] != NULL)
	{
		pcap_close(pds[i]);
		i++;
	}

}

inline int onnet(unsigned long int addr, struct ipnetwork *net)
{
  return ((addr & net -> netmask) == net -> netnumber);
}

int onnetlist(unsigned long int addr, struct ipnetwork *netlist)
{
    while(netlist!=NULL)
	{
	    if(onnet(addr, netlist))
		{
		    return 1;
		}
	    netlist = netlist -> next;
	}
    return 0;
}

void do_packet(u_char *usr, const struct pcap_pkthdr *h, const u_char *p)
{
  pthread_mutex_lock(&pt_lock);
  handle_frame((char *)p, h->len);
  pthread_mutex_unlock(&pt_lock);
}

/* 
 * Here goes explanation for using select() on BSD, it is from 
 * ethereal source code.
 * 
 * We don't want to do a "select()" on the pcap_t's file descriptor on
 * BSD (because "select()" doesn't work correctly on BPF devices on at
 * least some releases of some flavors of BSD)
 *
 * */
void *packet_loop(void *threadid)
{

  int *id_ptr, taskid;
  
  fd_set readmask;
  int pcap_fd;
  
  id_ptr = (int *) threadid;
  taskid = *id_ptr;

  /* BIG MESS HERE .. net to add proper #ifdef for varius OS */

  pcap_fd = pcap_fileno(pds[taskid]);
 
  /* infinite cycle and catch one packet with pcap_dispatch()
   * otherwise capturing will be blocked to other threads */
  while(1)
  {
#if !defined(_FreeBSD_) && !defined(_OpenBSD_)
	  FD_ZERO(&readmask);
	  FD_SET(pcap_fd, &readmask);
  
	  if(select(pcap_fd+1, &readmask, NULL, NULL, NULL)>0)
	  {
#endif
		  if(pcap_dispatch(pds[taskid], 1, do_packet, NULL) < 0) 
		  {
			  syslog(LOG_ERR, "pcap_dispatch: %s\n", pcap_geterr(pds[taskid]));
			  daemon_stop(0);
		  }
	  //syslog(LOG_DEBUG, "dev eth%i", taskid);
#if !defined(_FreeBSD_) && !defined(_OpenBSD_)
	  }
#endif
  }
  
}

/* real analysis of packet */
void handle_frame (unsigned char buf[], int length)
{
#if defined(_FreeBSD_) || defined(_OpenBSD_)  || defined(_Solaris_)
  static struct ip tmp_iphdr;
#else
  static struct iphdr tmp_iphdr;
#endif
  unsigned short srcport, dstport;
  struct tcphdr tmp_tcphdr;
  struct udphdr tmp_udphdr;
  struct icmp tmp_icmphdr;
  int found = 0;
  struct mon_host_struct *ptr;
  

  if(buf[12] * 256 + buf[13] == ETHERTYPE_IP)
    {
      memcpy (&tmp_iphdr, &(buf[14]), sizeof (tmp_iphdr));
      

      found = 0;
      if(cfg->hostlist) { /* we specified at least one hostlimit tag */
        /* if we don't monitor this IP, bail now - mk */
          for(ptr=cfg->hostlist;ptr && !found;ptr=ptr->next)
	  {

#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
		if(ptr->ipaddr == tmp_iphdr.ip_src.s_addr
	 	|| ptr->ipaddr == tmp_iphdr.ip_dst.s_addr)
#else
                if(ptr->ipaddr == tmp_iphdr.saddr
		|| ptr->ipaddr == tmp_iphdr.daddr)
#endif
                    found = 1;
          
          if(!found) {
                packets->ignored++;
                continue;
             }
          }
       }
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
      if((tmp_iphdr.ip_src.s_addr & cfg->ignoremask) ==
	 (tmp_iphdr.ip_dst.s_addr & cfg->ignoremask))
#else
      if((tmp_iphdr.saddr & cfg->ignoremask) ==
	 (tmp_iphdr.daddr & cfg->ignoremask))
#endif	      
        {
          packets->local++;
          return;
        }
      else
	{

#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	if(onnetlist(tmp_iphdr.ip_src.s_addr,cfg->ignorenet) ||
	     onnetlist(tmp_iphdr.ip_dst.s_addr, cfg->ignorenet))
#else
	if(onnetlist(tmp_iphdr.saddr,cfg->ignorenet) ||
             onnetlist(tmp_iphdr.daddr, cfg->ignorenet))
#endif		
		
            {
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
              if(!(onnetlist(tmp_iphdr.ip_src.s_addr,cfg->dontignore) ||
		   onnetlist(tmp_iphdr.ip_dst.s_addr, cfg->dontignore)))
#else
              if(!(onnetlist(tmp_iphdr.saddr,cfg->dontignore) ||
                   onnetlist(tmp_iphdr.daddr, cfg->dontignore)))
#endif		      
					    
                {
                  if(debug_level & DBG_IGNORE)
                    {
                      char tmp[18];
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
                      strcpy(tmp, intoa(tmp_iphdr.ip_src.s_addr));
#else
		      strcpy(tmp, intoa(tmp_iphdr.saddr));
#endif		      

#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
                      DEBUG(DBG_IGNORE, sprintf(dbg, "netignored: %s -> %s\n",
                            tmp,intoa(tmp_iphdr.ip_dst.s_addr)));
#else
                      DEBUG(DBG_IGNORE, sprintf(dbg, "netignored: %s -> %s\n",
		            tmp,intoa(tmp_iphdr.daddr)));
#endif		      
		      
                    }
                  packets->netignored++;
                  return;
                }
            }
	  packets->ip++;
	  
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	  switch(tmp_iphdr.ip_p)
#else
          switch(tmp_iphdr.protocol)
#endif	
	    {  
	    case IPPROTO_UDP:
	      packets->ip_udp++;
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	      memcpy (&tmp_udphdr, &buf[14 + tmp_iphdr.ip_hl * 4], sizeof (tmp_udphdr));
#else	      
	      memcpy (&tmp_udphdr, &buf[14 + tmp_iphdr.ihl * 4], sizeof (tmp_udphdr));
#endif	      

#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	      srcport = ntohs(tmp_udphdr.uh_sport);
	      dstport = ntohs(tmp_udphdr.uh_dport);
#else	      
	      srcport = ntohs(tmp_udphdr.source);
	      dstport = ntohs(tmp_udphdr.dest);
#endif
	      break;
	    case IPPROTO_TCP:
	      packets->ip_tcp++;
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	      memcpy (&tmp_tcphdr, &buf[14 + tmp_iphdr.ip_hl * 4], sizeof (tmp_tcphdr));
#else
	      memcpy (&tmp_tcphdr, &buf[14 + tmp_iphdr.ihl * 4], sizeof (tmp_tcphdr));
#endif	     

#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	      srcport = ntohs(tmp_tcphdr.th_sport);
              dstport = ntohs(tmp_tcphdr.th_dport);
#else	      	      
	      srcport = ntohs(tmp_tcphdr.source);
	      dstport = ntohs(tmp_tcphdr.dest);
#endif
	      break;
	    case IPPROTO_ICMP:
	      packets->ip_icmp++;
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	      memcpy (&tmp_icmphdr, &buf[14 + tmp_iphdr.ip_hl * 4], sizeof (tmp_icmphdr));
#else
	      memcpy (&tmp_icmphdr, &buf[14 + tmp_iphdr.ihl * 4], sizeof (tmp_icmphdr));
#endif	      
	      srcport = tmp_icmphdr.icmp_type;
	      dstport = tmp_icmphdr.icmp_code;

	      break;
	    default:
	      packets->ip_other++;
	      srcport = dstport = 0;
	      break;
	    }
#if defined(_FreeBSD_) || defined(_OpenBSD_) || defined(_Solaris_)
	  register_packet(tmp_iphdr.ip_src.s_addr,tmp_iphdr.ip_dst.s_addr,
			  ntohs(tmp_iphdr.ip_len), cfg->promisc->name);
#else

	  register_packet(tmp_iphdr.saddr,tmp_iphdr.daddr,
			  ntohs(tmp_iphdr.tot_len), cfg->promisc->name);

#endif
	}
    }
  else
    {
      /* ETH_P_ARP, ETH_P_RARP, ETH_P_IPX, etc. */
      packets -> ignored ++;
    }
}
